# /*
#  * Copyright (C) 2020, KylinSoft Co., Ltd.
#  *
#  * This program is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU General Public License as published by
#  * the Free Software Foundation, either version 3 of the License, or
#  * (at your option) any later version.
#  *
#  * This program is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  * GNU General Public License for more details.
#  *
#  * You should have received a copy of the GNU General Public License
#  * along with this program.  If not, see <https://www.gnu.org/licenses/>.
#  */

#!/usr/bin/env python

import gi
import traceback
import shutil
import os,subprocess
gi.require_version('OSTree', '1.0')
from gi.repository import GLib, Gio, OSTree,GObject
from SystemUpdater.Core.errors import *
from SystemUpdater.Core.enums import *
from SystemUpdater.constants import OFFLINE_REPO_POST,KYLIN_OS_NAME
from gettext import gettext as _
from SystemUpdater.Core.diskManager import check_free_space
import SystemUpdater.modules.self_upgrade as self_upgrade

RETRY_LIMIT = 5
DEFAULT_REPO_NAME = "kylin"

class BackendBaseOstree(GObject.GObject):    
    __gsignals__ = {
                    "transaction-done": (GObject.SignalFlags.RUN_FIRST,
                                         None,
                                         (GObject.TYPE_PYOBJECT,)),
                    "download-info": (GObject.SIGNAL_RUN_FIRST,
                                GObject.TYPE_NONE,
                                (   GObject.TYPE_INT,
                                    GObject.TYPE_INT,
                                    GObject.TYPE_ULONG,
                                    GObject.TYPE_ULONG,
                                    GObject.TYPE_ULONG
                                )),
                    "deploy-progress": (GObject.SIGNAL_RUN_FIRST,
                                GObject.TYPE_NONE,
                                (   GObject.TYPE_INT,
                                )),
                    }
    
    def __init__(self,parent,sysroot_path=None,osname=KYLIN_OS_NAME):
        GObject.GObject.__init__(self)
        self._async_progress = None
        if sysroot_path:
            sysroot_path = Gio.File.new_for_path(sysroot_path)
            
        self.sysroot = OSTree.Sysroot.new(sysroot_path)
        self.sysroot.set_mount_namespace_in_use()
        self.sysroot.initialize()
        self.sysroot.load()

        booted_deployment = self.sysroot.get_booted_deployment()
        if booted_deployment:
            self.osname = booted_deployment.get_osname()
        else:
            logging.warning("Booted deployment is None and use to defalut os name(%s)...",osname)
            self.osname = osname

        self.cur_deployment = self.sysroot.get_merge_deployment(self.osname)
        if not self.cur_deployment:
            logging.error("Now system deployment is None...")
            
        self.origin = self.cur_deployment.get_origin()

        # 加载插件机制
        # 获取当前运行程序的插件目录位置
        # self.plugins_dir = os.path.join(os.getcwd(),"SystemUpdater/plugins")
        # # 基础插件 所有插件的位置
        # test = os.path.join(self.plugins_dir,"base")
        # self.plugin_base = PluginBase(package='SystemUpdater.plugins',searchpath=[os.path.join(self.plugins_dir,"base")])

    def run(self,trans=None):
        logging.info("Processing transaction")
        # self._run_transaction_idle(trans)
        GLib.idle_add(self._run_transaction_idle,trans)

    def _run_transaction_idle(self,trans):
        """Run the transaction"""
        try:
            self._async_progress = OSTree.AsyncProgress.new()
            self._async_progress.connect("changed",self._on_async_progress)
            self._try_lock()
            self._run_transaction(self.sysroot,trans)
        except GLib.Error as error:
            logging.error(str(error))
            trans.success = False
            self._parse_ostree_error(error,trans)
        except UpdateBaseError as excep:
            trans.success = False
            trans.error = excep.code
        except Exception as excep:
            logging.error(str(excep))
            trans.success = False

            tbk = traceback.format_exc()
            logging.error(str(tbk))
        else:
            trans.success = True
        
        finally:
            self._try_unlock()
            self._async_progress.finish()
            logging.info("Finished transaction")

            self._emit_transaction_done(trans)

        return False

    def _run_transaction(self,sysroot,trans):
        """This method needs to be implemented by the backends."""
        pass
        # raise errors.TransactionFailed(enums.ERROR_NOT_SUPPORTED)

    def _try_lock(self):
        ret,out_acquired = self.sysroot.try_lock()
        if not out_acquired:
            raise UpdateBaseError(ERROR_OTHER_PROGRESS_OCCUPATION)
    
    def _try_unlock(self):
        self.sysroot.unlock()

    def _emit_transaction_done(self,trans):
        """Emit the transaction-done signal.

        Keyword argument:
        trans -- the finished transaction
        """
        # logging.debug("Emitting transaction-done: %d",)
        self.emit("transaction-done",trans)

    def _on_async_progress(self,obj):
        try:
            start_time = obj.get_uint64("start-time")
            bytes_transferred = obj.get_uint64("bytes-transferred") 

            fetched = obj.get_uint("fetched")
            requested = obj.get_uint("requested")

            # 下载速度
            elapsed_secs=(GLib.get_monotonic_time()-start_time)/GLib.USEC_PER_SEC
            bytes_sec = int(bytes_transferred / elapsed_secs)

            if requested <= 2:
                return
            
            self.emit("download-info",fetched,requested,bytes_transferred,0,bytes_sec)
        except Exception as excep:
            tbk = traceback.format_exc()
            logging.error(str(tbk))

    def _parse_ostree_error(self,error,trans):
        if trans.role == ROLE_DEPLOY:
            trans.error = ERROR_NOT_DEPLOYMENT
        else:
            if error.code in (0,37):
                trans.error = ERROR_NETWORK_FAILED
            #取消
            elif error.code == 19:
                trans.cancellable.reset()
                trans.error = EXIT_CANCELLED
            elif error.code == 1:
                # 在远程分支 未找到 对应的 分支 
                if "Server returned status 404" in error.message:
                    trans.error = ERROR_PUSH_BRANCH_EXCEPTION
                # 未找到 对应的仓库信息 
                elif "Remote" in error.message and "not found" in error.message:
                    trans.success = True
    
    def _open_offline_repo(self):
        # 检查是否存在离线仓库
        offline_flag = OFFLINE_REPO_POST + "/config"
        if os.path.exists(offline_flag):
            g_file= Gio.File.new_for_path(OFFLINE_REPO_POST)
            offline_repo = OSTree.Repo.new(g_file)
            ret = offline_repo.open()
            return offline_repo
        else:
            return None

    def register_formatter(self, name, formatter):
        """A function a plugin can use to register a formatter."""
        self.formatters[name] = formatter

    def _load_plugins(self,identifier):
        # 每次formatters将被刷新
        self.formatters = {}
        plugin_dir = os.path.join(self.plugins_dir,identifier)
        plugins_sources = self.plugin_base.make_plugin_source(searchpath=[plugin_dir],identifier=identifier)

        # 初始化插件
        for plugin_name in plugins_sources.list_plugins():
            plugin = plugins_sources.load_plugin(plugin_name)
            plugin.setup(self)
        return self.formatters