import subprocess
import logging
from typing import List

from .manifest import (
    PostAction,
)

import subprocess
import logging

def _execute_command(hook: str):
    """
    执行 shell 命令 hook。
    - 返回 (True, None) 表示执行成功且无错误输出。
    - 返回 (False, stderr) 表示执行失败或有错误输出。
    """
    if not hook:
        logging.info("No hook to execute.")
        return True, None  # 空钩子不视为错误

    try:
        logging.info(f"Executing hook: {hook}")
        result = subprocess.run(hook, shell=True, text=True, capture_output=True)

        if result.stdout:
            logging.info(f"Stdout: {result.stdout.strip()}")

        stderr = result.stderr.strip() if result.stderr else None
        if stderr:
            logging.warning(f"Stderr: {stderr}")

        if result.returncode == 0 and not stderr:
            return True, None
        else:
            return False, stderr or "Non-zero return code without stderr."

    except Exception as ex:
        logging.error(f"Failed to execute hook '{hook}': {str(ex)}")
        return False, str(ex)

def execute_actions(actions: List[PostAction]):
    if not actions:
        logging.info("No hooks to execute.")
        return actions

    for action in actions:
        command = f"{action.command} {action.value}".strip()
        action.success,action.error_desc = _execute_command(command)

    return actions

def execute_hooks(hook_paths: List[str]):
    if not hook_paths:
        logging.info("No hooks to execute.")
        return

    for hook in hook_paths:
        try:
            logging.info(f"Executing hook: {hook}")
            # 使用 shell=True 可以让我们执行完整命令字符串（如 "systemctl restart xxx"）
            result = subprocess.run(hook, shell=True, check=True, text=True, capture_output=True)
            if result.stdout:
                logging.info(f"Output from '{hook}': {result.stdout.strip()}")
            if result.stderr:
                logging.warning(f"Stderr from '{hook}': {result.stderr.strip()}")
        except subprocess.CalledProcessError as e:
            logging.error(f"Hook execution failed for '{hook}' (exit code {e.returncode}). Error: {e.stderr.strip()}")
        except Exception as ex:
            logging.error(f"Unexpected error while executing '{hook}': {str(ex)}")
