# repo_checkout.py
import os
import logging
from typing import List
from gi.repository import GLib, OSTree

def checkout_files_partly(repo_obj, trans, pull_dirs, pull_files, put_dir) -> List[str]:
    """
    部分检出指定目录和文件到目标目录，并返回成功检出的相对路径列表。
    """
    success_files: List[str] = []

    if not (pull_dirs or pull_files):
        logging.info("No files or directories to checkout.")
        return success_files

    if not os.path.exists(put_dir):
        os.makedirs(put_dir, exist_ok=True)

    g_options = {
        "n-network-retries": GLib.Variant("u", 5),
        "flags": GLib.Variant("i", OSTree.RepoPullFlags.NONE),
        "depth": GLib.Variant("i", 1),
        "subdirs": GLib.Variant("as", pull_dirs + pull_files),
        "refs": GLib.Variant("as", [trans.available_refs])
    }

    try:
        repo_obj.pull_with_options(trans.available_repo, GLib.Variant("a{sv}", g_options), None, None)
    except Exception as ex:
        logging.error("Failed to pull content: %s", ex)
        return success_files

    postinst_opts = OSTree.RepoCheckoutAtOptions()
    postinst_opts.overwrite_mode = OSTree.RepoCheckoutOverwriteMode(1)

    def _checkout_subpath(subpath: str, file_dir: str) -> bool:
        """内部工具函数：尝试检出单个路径，成功返回True，失败返回False"""
        postinst_opts.subpath = subpath
        os.makedirs(file_dir, exist_ok=True)
        try:
            repo_obj.checkout_at(postinst_opts, 0, file_dir, trans.new_revision, None)
            return True
        except Exception as excep:
            logging.warning("Checkout failed for '%s': %s", subpath, excep)
            return False

    # Checkout 目录
    if pull_dirs:
        logging.info("Start checking out %d directories to %s", len(pull_dirs), put_dir)
        for dir_path in pull_dirs:
            file_dir = os.path.join(put_dir, dir_path.lstrip('/'))
            if _checkout_subpath(dir_path,file_dir):
                success_files.append(dir_path)

    # Checkout 文件
    if pull_files:
        logging.info("Start checking out %d files to %s", len(pull_files), put_dir)
        for file_path in pull_files:
            file_dir = os.path.join(put_dir, os.path.dirname(file_path).lstrip('/'))
            if _checkout_subpath(file_path,file_dir):
                success_files.append(file_path)

    return success_files

