# update.py
import os
import json
import shutil
import logging
from typing import Dict,List,Tuple,Optional
from pydantic import BaseModel

from .partial_checkout import checkout_files_partly
from .dpkg import get_versions_from_status
from .hooks import execute_actions
from .config import VERSIONS_DIR, MANIFESTS_DIR, CURRENT_LINK
from .manifest import (
    PostAction,
    ComponentManifest,
    parse_manifests,
    get_use_dpkg_list_packages,
    collect_installed_files_from_manifests,
    get_move_files_from_manifests,
    get_all_post_hooks,
    handle_move_operations,
    get_remapped_files
)

VERSION_MANIFEST_FILE = "version-manifests.json"
DPKG_STATUS_PATH = "/var/lib/dpkg/status"
CHECKOUT_FILES_LIMIT = 10

class VersionConfig(BaseModel):
    pull_ref: str = None               # 拉取来源分支
    commit_id: str = None              # 当前版本commit id
    root_dir: str = None               # 版本根目录
    packages: Dict[str, str] = {}
    moved_files: Dict[str, str] = {}
    executed_hooks: List[PostAction] = []     # 所有更新组件自升级完成后 需要做的处理，例如重启服务
    manifests: Dict[str, ComponentManifest] = {}  # 每个组件的manifest描述

def get_current_version() -> Tuple[str, VersionConfig]:
    commit_id = None
    version_config = VersionConfig()

    if os.path.islink(CURRENT_LINK):
        try:
            commit_id = os.path.basename(os.readlink(CURRENT_LINK))
        except OSError as e:
            logging.warning(f"Failed to read current symlink: {e}")
            return None, version_config

    if commit_id:
        version_manifest_path = os.path.join(CURRENT_LINK, VERSION_MANIFEST_FILE)
        if os.path.exists(version_manifest_path):
            try:
                with open(version_manifest_path, 'r') as f:
                    data = json.load(f)
                version_config = VersionConfig(**data)
            except (OSError, json.JSONDecodeError) as e:
                logging.error(f"Failed to load version manifest: {e}")

    return commit_id, version_config

def _clean_old_versions(new_version_id: str, old_version_id: str):
    for version_name in os.listdir(VERSIONS_DIR):
        version_path = os.path.join(VERSIONS_DIR, version_name)
        if version_name not in {new_version_id, old_version_id} and os.path.isdir(version_path):
            logging.info("Deleting outdated version: %s", version_name)
            shutil.rmtree(version_path)

def _clean_old_config(new_moved_files: Dict[str, str], old_moved_files: Dict[str, str]):
    """
    清理旧版本遗留的配置文件。
    
    规则：
    - 如果旧版本迁移的源文件在新版本中已不存在，删除对应的目标文件。
    - 如果源文件存在，但目标路径变化，删除旧目标文件。
    - 如果源文件和目标路径均未变化，保留目标文件。
    """
    for src_file, old_dst_file in old_moved_files.items():
        new_dst_file = new_moved_files.get(src_file)

        if new_dst_file is None:
            # 新版本已无此迁移项，直接删除旧文件
            if os.path.exists(old_dst_file):
                logging.info("Deleting outdated config file: %s", old_dst_file)
                os.remove(old_dst_file)
        elif new_dst_file != old_dst_file:
            # 迁移路径发生变化，删除旧目标文件
            if os.path.exists(old_dst_file):
                logging.info("Deleting outdated config file due to path change: %s",old_dst_file)
                os.remove(old_dst_file)
        else:
            # 迁移路径未变化，保留文件
            logging.info("Keeping valid config file: %s", old_dst_file)

def activate_new_version(version_path: str):
    if not os.path.isdir(version_path):
        logging.error("Invalid version path: %s", version_path)
        return

    # 最后建立软链接 进行激活运行版本
    if os.path.exists(CURRENT_LINK):
        os.remove(CURRENT_LINK)
    os.symlink(version_path, CURRENT_LINK)

    logging.info("Activated version: %s", version_path)

def _should_upgrade(new_version_id: str) -> bool:
    version_path = os.path.join(VERSIONS_DIR, new_version_id)

    if not os.path.isdir(version_path):
        logging.info("Upgrade needed: version path does not exist: %s", version_path)
        return True

    if os.path.islink(CURRENT_LINK):
        try:
            current_commit_id = os.path.basename(os.readlink(CURRENT_LINK))
            if current_commit_id == new_version_id:
                logging.info("No upgrade needed: already running version %s", new_version_id)
                return False
        except OSError as e:
            logging.warning(f"Failed to read current symlink: {e}")

    # 目录存在但未激活，删除重新拉取
    logging.info("Upgrade needed: existing version is not active")
    shutil.rmtree(version_path, ignore_errors=True)
    return True

def add_new_version(repo_obj, trans,version_root) -> VersionConfig:
    version_config = VersionConfig(
        pull_ref=trans.available_refs,
        commit_id=trans.new_revision,
        root_dir=version_root,
    )

    logging.info("Checking out update manifests...")
    checkout_files_partly(repo_obj, trans, [MANIFESTS_DIR], [], version_root)
    version_config.manifests = parse_manifests(version_root)
    if not version_config.manifests:
        logging.warning("No manifests found, aborting self upgrade add new version addition.")
        return None

    logging.info("Checking out dpkg list and status...")
    dpkg_files = get_use_dpkg_list_packages(version_config.manifests)
    dpkg_files.append(DPKG_STATUS_PATH) # 追加拉取status文件
    checkout_files_partly(repo_obj, trans, [], dpkg_files, version_root)

    # 解析status文件，记录每个软件包的版本信息
    packages_list = list(version_config.manifests.keys())
    status_file = os.path.join(version_root, DPKG_STATUS_PATH.lstrip('/'))
    version_config.packages = get_versions_from_status(packages_list,status_file)

    logging.info("Starting checkout package files:%s",version_config.packages)
    all_files = collect_installed_files_from_manifests(version_config.manifests)
    success_files = checkout_files_partly(repo_obj, trans, [], all_files, version_root)
    if len(success_files) < CHECKOUT_FILES_LIMIT:
        logging.warning("Success files(%s) less than 10 files checked out, aborting...",len(success_files))
        return None

    return version_config


def run_priority_upgrade(repo_obj, trans):
    new_version_id = trans.new_revision
    if not _should_upgrade(new_version_id):
        return False

    old_version_id,old_version_config = get_current_version()

    # 创建/opt/update-module/versions/下面根目录
    version_root = os.path.join(VERSIONS_DIR, new_version_id)
    os.makedirs(version_root, exist_ok=True)

    # 下载优先升级版本内容
    version_config = add_new_version(repo_obj, trans,version_root)
    if version_config is None:
        # 优先升级失败时 删除掉下载的内容
        shutil.rmtree(version_root, ignore_errors=True)
        return False

    # 进行规则映射，将配置转移到/opt规则下面
    remapped_files = get_remapped_files(version_config.manifests)
    has_moved = handle_move_operations(remapped_files,version_root)
    version_config.moved_files.update(has_moved)
    
    # 处理需要复制文件到指定的位置
    move_files = get_move_files_from_manifests(version_config.manifests)
    has_moved = handle_move_operations(move_files,version_root)
    version_config.moved_files.update(has_moved)

    # 最后进行激活目录，保证原子性
    activate_new_version(version_root)

    # 删除旧版本与清除旧配置
    _clean_old_versions(new_version_id, old_version_id)
    _clean_old_config(version_config.moved_files, old_version_config.moved_files)

    # 执行自更新后的服务
    post_hooks = get_all_post_hooks(version_config.manifests)
    for action in post_hooks:
        if MANIFESTS_DIR in action.command:
            action.command = os.path.join(version_root, action.command.lstrip('/'))
    version_config.executed_hooks = execute_actions(post_hooks)

    # 导出配置到本地
    manifest_path = os.path.join(version_root, VERSION_MANIFEST_FILE)
    with open(manifest_path, "w") as f:
        json.dump(json.loads(version_config.json()), f, indent=4)

    return True